import sys
from pathlib import Path
from tempfile import TemporaryDirectory
from time import sleep
from typing import Optional


class AltTemporaryDirectory:
    def __init__(self, directory: Optional[str] = None):
        self.tmpdir = TemporaryDirectory()
        self._extended_path = False
        self._directory = directory or ""
        name = str(Path(self.tmpdir.name) / self._directory)
        if name not in sys.path:
            self._extended_path = True
            sys.path.append(name)

    def __enter__(self):
        return self.tmpdir.name

    def cleanup(self, cnt=0):
        if self._extended_path:
            name = str(Path(self.tmpdir.name) / self._directory)
            if name in sys.path:
                sys.path.remove(name)
        if cnt >= 5:  # pragma: no cover
            raise RuntimeError("Could not delete TemporaryDirectory!")
        try:
            self.tmpdir.cleanup()
        except IOError:  # pragma: no cov_4_nix
            sleep(1)
            self.cleanup(cnt + 1)

    def __exit__(self, exc, value, tb):
        self.cleanup()



from pathlib import Path

from cruft import exceptions


def test_invalid_cookiecutter_repository():
    instance = exceptions.InvalidCookiecutterRepository(".")
    assert isinstance(instance, exceptions.CruftError)
    assert instance.cookiecutter_repo == "."


def test_unable_to_find_cookiecutter_template():
    instance = exceptions.UnableToFindCookiecutterTemplate(".")
    assert instance.directory == "."
    assert isinstance(instance, exceptions.CruftError)


def test_unable_to_find_cookiecutter_template_path():
    instance = exceptions.UnableToFindCookiecutterTemplate(Path("."))
    assert instance.directory == "."
    assert isinstance(instance, exceptions.CruftError)


def test_no_cruft():
    instance = exceptions.NoCruftFound(".")
    assert instance.directory == "."
    assert isinstance(instance, exceptions.CruftError)


def test_cruft_already_present():
    instance = exceptions.CruftAlreadyPresent(".")
    assert instance.file_location == "."
    assert isinstance(instance, exceptions.CruftError)



from pathlib import Path
from textwrap import dedent

import pytest

from cruft import exceptions
from cruft._commands import utils


def test_get_diff_with_add(tmp_path: Path):
    repo0 = tmp_path / "repo0"
    repo1 = tmp_path / "repo1"

    repo0.mkdir()
    repo1.mkdir()

    (repo1 / "file").touch()

    diff = utils.diff.get_diff(repo0, repo1)

    assert diff.startswith("diff --git upstream-template-old/file upstream-template-new/file")


def test_get_diff_with_delete(tmp_path: Path):
    repo0 = tmp_path / "repo0"
    repo1 = tmp_path / "repo1"

    repo0.mkdir()
    repo1.mkdir()

    (repo0 / "file").touch()

    diff = utils.diff.get_diff(repo0, repo1)

    assert diff.startswith("diff --git upstream-template-old/file upstream-template-new/file")


def test_get_diff_with_unicode(project_dir):
    with pytest.raises(exceptions.ChangesetUnicodeError):
        utils.diff.get_diff(
            Path(project_dir, "tests", "testdata", "unicode-data").absolute(),
            Path(project_dir, "tests", "testdata", "non-unicode-data").absolute(),
        )


def test_remove_paths_with_pathlib(tmp_path: Path):
    repo0 = tmp_path / "repo0"
    (repo0 / "tests").mkdir(parents=True)

    (repo0 / "file").touch()
    (repo0 / "tests" / "test0.py").touch()
    (repo0 / "tests" / "test1.py").touch()

    utils.generate._remove_paths(repo0, {Path("tests/test0.py")})
    assert not (repo0 / "tests" / "test0.py").exists()
    assert (repo0 / "tests" / "test1.py").exists()


def test_remove_paths_with_string(tmp_path: Path):
    repo0 = tmp_path / "repo0"
    (repo0 / "tests").mkdir(parents=True)

    (repo0 / "file").touch()
    (repo0 / "tests" / "test0.py").touch()
    (repo0 / "tests" / "test1.py").touch()

    utils.generate._remove_paths(repo0, {"tests/test0.py"})
    assert not (repo0 / "tests" / "test0.py").exists()
    assert (repo0 / "tests" / "test1.py").exists()


def test_remove_paths_with_glob_pattern(tmp_path: Path):
    repo0 = tmp_path / "repo0"
    (repo0 / "tests").mkdir(parents=True)

    (repo0 / "file").touch()
    (repo0 / "tests" / "test0.py").touch()
    (repo0 / "tests" / "test1.py").touch()

    utils.generate._remove_paths(repo0, {"tests/*"})
    assert not (repo0 / "tests" / "test0.py").exists()
    assert not (repo0 / "tests" / "test1.py").exists()


def test_remove_paths_folder(tmp_path: Path):
    repo0 = tmp_path / "repo0"
    (repo0 / "tests").mkdir(parents=True)

    (repo0 / "file").touch()
    (repo0 / "tests" / "test0.py").touch()
    (repo0 / "tests" / "test1.py").touch()

    utils.generate._remove_paths(repo0, {"tests", "file"})
    assert not (repo0 / "tests").exists()
    assert not (repo0 / "file").exists()


def test_remove_paths_with_glob_pattern_and_string(tmp_path: Path):
    repo0 = tmp_path / "repo0"
    (repo0 / "tests").mkdir(parents=True)

    (repo0 / "file").touch()
    (repo0 / "tests" / "test0.py").touch()
    (repo0 / "tests" / "test1.py").touch()

    with pytest.warns(None) as warn_record:
        utils.generate._remove_paths(repo0, {5})  # type: ignore
    assert len(warn_record) != 0, "a warning should have been called as typing was off"
    assert (repo0 / "tests" / "test0.py").exists()
    assert (repo0 / "tests" / "test1.py").exists()


def test_warn_if_cant_read_pyproject_toml(monkeypatch):
    monkeypatch.setattr(utils.generate, "tomllib", None)
    with pytest.warns(UserWarning, match="`toml` package is not installed"):
        utils.generate._get_skip_paths({}, Path(__file__))


def test_get_extra_context_from_file():
    extra_context_file = Path(__file__).parent / "testdata" / "unicode-data" / "extra_context.json"
    extra_context = utils.cookiecutter.get_extra_context_from_file(extra_context_file)
    assert extra_context == {"project": "CRUFT-TEST-PROJECT"}


def test_remove_paths_with_get_skip_paths_from_cruft_state_and_pyproject_toml(tmp_path: Path):
    repo0 = tmp_path / "repo0"
    repo0.mkdir(parents=True)

    (repo0 / "file0").touch()
    (repo0 / "file1").touch()

    (repo0 / "tests").mkdir(parents=True)
    (repo0 / "tests" / "test0.py").touch()
    (repo0 / "tests" / "test1.py").touch()

    (repo0 / "package" / "__pycache__").mkdir(parents=True)
    (repo0 / "package" / "__init__.py").touch()
    (repo0 / "package" / "test0.py").touch()
    (repo0 / "package" / "test1.py").touch()
    (repo0 / "package" / "__pycache__" / "test0.cpython-38.pyc").touch()
    (repo0 / "package" / "__pycache__" / "test1.cpython-38.pyc").touch()
    (repo0 / "package" / "module" / "__pycache__").mkdir(parents=True)
    (repo0 / "package" / "module" / "__init__.py").touch()
    (repo0 / "package" / "module" / "__pycache__" / "test0.cpython-38.pyc").touch()
    (repo0 / "package" / "module" / "test0.py").touch()

    (repo0 / ".mypy_cache").mkdir(parents=True)
    (repo0 / ".mypy_cache" / "test0.json").touch()
    (repo0 / ".ruff_cache").mkdir(parents=True)
    (repo0 / ".ruff_cache" / "test0.json").touch()

    (repo0 / ".git").mkdir(parents=True)
    (repo0 / ".git" / "file").touch()
    (repo0 / ".git" / "tests").mkdir(parents=True)
    (repo0 / ".git" / "tests" / "test0.py").touch()

    pyproject = repo0 / "pyproject.toml"
    with pyproject.open("w") as f:
        f.write(
            dedent(
                """
            [tool.cruft]
            skip = [
                "file1",
                "tests/*",
                "**/__pycache__/",
                "**/__init__.py",
                "*_cache/",
            ]
            """
            )
        )

    paths_to_remove = utils.generate._get_skip_paths({"skip": [".git"]}, pyproject)
    utils.generate._remove_paths(repo0, paths_to_remove)

    assert (repo0 / "file0").exists()
    assert not (repo0 / "file1").exists()

    assert not (repo0 / ".git").exists()

    assert (repo0 / "tests").exists()
    assert not (repo0 / "tests" / "test0.py").exists()
    assert not (repo0 / "tests" / "test0.py").exists()

    assert not (repo0 / "package" / "__pycache__").exists()
    assert (repo0 / "package").exists()

    assert not (repo0 / "package" / "module" / "__pycache__").exists()
    assert (repo0 / "package" / "module").exists()

    assert not (repo0 / ".mypy_cache").exists()
    assert not (repo0 / ".ruff_cache").exists()



import json
import os
import re
import sys
from pathlib import Path
from subprocess import run

import pytest
from examples import verify_and_test_examples
from git import Repo

import cruft
from cruft import exceptions
from cruft._commands import utils


def test_invalid_cookiecutter_repo(tmpdir):
    with pytest.raises(exceptions.InvalidCookiecutterRepository):
        cruft.create("DNE", Path(tmpdir))


def test_invalid_cookiecutter_reference(tmpdir):
    with pytest.raises(exceptions.InvalidCookiecutterRepository):
        cruft.create("https://github.com/cruft/cookiecutter-test", Path(tmpdir), checkout="DNE")


def test_no_cookiecutter_dir(tmpdir):
    with pytest.raises(exceptions.UnableToFindCookiecutterTemplate):
        cruft.create("https://github.com/cruft/cookiecutter-test", Path(tmpdir))


def test_create_examples(tmpdir):
    tmpdir.chdir()
    verify_and_test_examples(cruft.create)


def test_check_examples(tmpdir, project_dir):
    tmpdir.chdir()
    with pytest.raises(exceptions.NoCruftFound):
        verify_and_test_examples(cruft.check)

    os.chdir(project_dir)
    verify_and_test_examples(cruft.check)


def test_create_with_skips(tmpdir):
    tmpdir.chdir()
    skips = ["setup.cfg"]
    cruft.create("https://github.com/cruft/cookiecutter-python", Path(tmpdir), skip=skips)

    assert json.load((tmpdir / "python_project_name" / ".cruft.json").open("r"))["skip"] == skips


@pytest.mark.parametrize("value", ["main", None])
def test_create_stores_checkout_value(value, tmpdir):
    tmpdir.chdir()

    cruft.create("https://github.com/cruft/cookiecutter-python", Path(tmpdir), checkout=value)

    assert (
        json.load((tmpdir / "python_project_name" / ".cruft.json").open("r"))["checkout"] == value
    )


@pytest.mark.parametrize("value", ["main", None])
def test_link_stores_checkout_value(value, tmpdir):
    project_dir = Path(tmpdir)
    cruft.link(
        "https://github.com/cruft/cookiecutter-python",
        project_dir=project_dir,
        checkout=value,
    )

    assert json.load(utils.cruft.get_cruft_file(project_dir).open("r"))["checkout"] == value


@pytest.mark.parametrize("value", ["main", None])
def test_update_stores_checkout_value(value, tmpdir):
    tmpdir.chdir()
    cruft.create(
        "https://github.com/cruft/cookiecutter-python",
        Path(tmpdir),
        checkout="ea8f733f85e7089df338d41ace199d3f4d397e29",
    )
    project_dir = tmpdir / "python_project_name"

    cruft.update(Path(project_dir), checkout=value)

    assert json.load((project_dir / ".cruft.json").open("r"))["checkout"] == value


def test_update_and_check_real_repo(tmpdir):
    tmpdir.chdir()
    repo = Repo.clone_from("https://github.com/cruft/cruft", str(tmpdir))
    repo.head.reset(commit="86a6e6beda8095690414ff7652c15b7ae36e6128", working_tree=True)
    with open(os.path.join(tmpdir, ".cruft.json")) as cruft_file:
        cruft_state = json.load(cruft_file)
        cruft_state["skip"] = ["cruft/__init__.py", "tests"]
    with open(os.path.join(tmpdir, ".cruft.json"), "w") as cruft_file:
        json.dump(cruft_state, cruft_file)
    repo_dir = Path(tmpdir)
    assert not cruft.check(repo_dir)
    # Update should fail since we have an unclean git repo
    assert not cruft.update(repo_dir)
    # Commit the changes so that the repo is clean
    run(
        [
            "git",
            "-c",
            "user.name='test'",
            "-c",
            "user.email='user@test.com'",
            "commit",
            "-am",
            "test",
        ],
        cwd=repo_dir,
    )
    assert cruft.update(repo_dir, skip_apply_ask=True)


def test_update_allows_untracked_files_option(tmpdir):
    tmpdir.chdir()
    Repo.clone_from("https://github.com/cruft/cruft", str(tmpdir))
    with open(os.path.join(tmpdir, "untracked.txt"), "w") as new_file:
        new_file.write("hello, world!\n")
    repo_dir = Path(tmpdir)
    # update should fail since repo is now unclean (has a tracked file)
    assert not cruft.update(repo_dir)
    # update should work if allow_untracked_files is True
    assert cruft.update(repo_dir, allow_untracked_files=True)


def test_update_locally_cloned_template(tmpdir):
    tmpdir.chdir()
    repo_dir = Path(tmpdir) / "cruft"
    template_dir = Path(tmpdir) / "template"
    cookiecutter_repo = "https://github.com/cruft/cookiecutter-python"
    Repo.clone_from("https://github.com/cruft/cruft", repo_dir)
    Repo.clone_from(cookiecutter_repo, template_dir)
    with open(os.path.join(repo_dir, "untracked.txt"), "w") as new_file:
        new_file.write("hello, world!\n")
    # update should fail since repo is now unclean (has a tracked file)
    assert not cruft.update(repo_dir, template_path=template_dir)
    # update should work if allow_untracked_files is True
    assert cruft.update(repo_dir, template_path=template_dir, allow_untracked_files=True)
    assert (
        json.loads(utils.cruft.get_cruft_file(repo_dir).read_text())["template"]
        == cookiecutter_repo
    )


def test_relative_repo_check(tmpdir):
    tmpdir.chdir()
    temp_dir = Path(tmpdir)
    Repo.clone_from("https://github.com/cruft/cookiecutter-test", str(temp_dir / "cc"))
    project_dir = cruft.create("./cc", output_dir=str(temp_dir / "output"), directory="dir")
    assert cruft.check(project_dir)


def test_update_examples(project_dir, tmpdir):
    tmpdir.chdir()
    with pytest.raises(exceptions.NoCruftFound):
        verify_and_test_examples(cruft.update)

    os.chdir(project_dir)
    verify_and_test_examples(cruft.update)


def test_link_examples(project_dir, tmpdir):
    os.chdir(project_dir)
    with pytest.raises(exceptions.CruftAlreadyPresent):
        verify_and_test_examples(cruft.link)

    tmpdir.chdir()
    Repo.clone_from("https://github.com/cruft/cruft", str(tmpdir))
    os.remove(os.path.join(tmpdir, ".cruft.json"))
    verify_and_test_examples(cruft.link)


def test_directory_and_checkout(tmpdir):
    output_path = cruft.create(
        "https://github.com/cruft/cookiecutter-test",
        output_dir=Path(tmpdir),
        directory="dir",
        checkout="initial",
    )
    cruft_file = utils.cruft.get_cruft_file(output_path)
    assert cruft_file.exists()
    assert cruft.check(output_path, checkout="initial")
    assert not cruft.check(output_path, checkout="updated")
    assert cruft.update(output_path, checkout="updated")
    assert cruft.check(output_path, checkout="updated")
    cruft_file.unlink()
    assert not cruft_file.exists()
    assert cruft.link(
        "https://github.com/cruft/cookiecutter-test",
        project_dir=output_path,
        directory="dir",
        checkout="updated",
    )
    assert cruft.check(output_path, checkout="updated")
    # Add checks for strictness where main is an older
    # version than updated
    assert not cruft.check(output_path, strict=True)
    assert cruft.check(output_path, strict=False)


@pytest.mark.parametrize(
    "exit_code,isatty,expect_reproducible_diff,expected_return_value",
    [
        (False, False, True, True),  # $ cruft diff | cat
        (False, True, False, True),  # $ cruft diff
        (True, False, True, False),  # $ cruft diff --exit-code | cat
        (True, True, False, False),  # $ cruft diff --exit-code
    ],
)
def test_diff_has_diff(
    exit_code, isatty, expect_reproducible_diff, expected_return_value, capfd, mocker, tmpdir
):
    mocker.patch.object(sys.stdout, "isatty", return_value=isatty)

    project_dir = cruft.create(
        "https://github.com/cruft/cookiecutter-test", Path(tmpdir), directory="dir", checkout="diff"
    )
    (project_dir / "file0").write_text("new content 0\n")
    (project_dir / "dir0/file1").write_text("new content 1\n")
    (project_dir / "dir0/file2").unlink()

    assert cruft.diff(project_dir, exit_code=exit_code) == expected_return_value

    captured = capfd.readouterr()
    stdout = captured.out
    stderr = captured.err

    assert stderr == ""

    expected_output = """diff --git upstream-template-old{tmpdir}/dir0/file1 upstream-template-new{tmpdir}/dir0/file1
index eaae237..ac3e272 100644
--- upstream-template-old{tmpdir}/dir0/file1
+++ upstream-template-new{tmpdir}/dir0/file1
@@ -1 +1 @@
-new content 1
+content1
diff --git upstream-template-old{tmpdir}/file0 upstream-template-new{tmpdir}/file0
index be6a56b..1fc03a9 100644
--- upstream-template-old{tmpdir}/file0
+++ upstream-template-new{tmpdir}/file0
@@ -1 +1 @@
-new content 0
+content0
"""
    expected_output_regex = re.escape(expected_output)
    expected_output_regex = expected_output_regex.replace(r"\{tmpdir\}", r"([^\n]*)")
    expected_output_regex = rf"^{expected_output_regex}$"

    match = re.search(expected_output_regex, stdout, re.MULTILINE)
    assert match is not None

    if expect_reproducible_diff:
        # If the output is not displayed to the user (for example when piping the result
        # of the "cruft diff" command) or if the user requested an exit code, we must make
        # sure the absolute path to the temporary directory does not appear in the diff
        # because the user might want to process the output.
        # Conversely, when the output is supposed to be displayed to the user directly (e.g.
        # when running "cruft diff" command directly in a terminal), absolute path to the
        # actual files on disk may be displayed because git diff command is called directly
        # without reprocessing by cruft. This delegates diff coloring and paging to git which
        # improves user experience. As far as I know, there is no way to ask git diff to not
        # display this path.
        assert set(match.groups()) == {""}


@pytest.mark.parametrize("exit_code", [(False,), (True,)])
def test_diff_no_diff(exit_code, capfd, mocker, tmpdir):
    project_dir = cruft.create(
        "https://github.com/cruft/cookiecutter-test", Path(tmpdir), directory="dir", checkout="diff"
    )

    assert cruft.diff(project_dir, exit_code=exit_code) is True

    captured = capfd.readouterr()
    stdout = captured.out
    stderr = captured.err

    assert stdout == ""
    assert stderr == ""


def test_diff_checkout(capfd, tmpdir):
    project_dir = cruft.create(
        "https://github.com/cruft/cookiecutter-test",
        Path(tmpdir),
        directory="dir",
        checkout="main",
    )

    assert cruft.diff(project_dir, exit_code=True, checkout="updated") is False

    captured = capfd.readouterr()
    stdout = captured.out
    stderr = captured.err

    assert stderr == ""
    assert "--- upstream-template-old/README.md" in stdout
    assert "+++ upstream-template-new/README.md" in stdout
    assert "+Updated again" in stdout
    assert "-Updated" in stdout


def test_diff_git_subdir(capfd, tmpdir):
    tmpdir.chdir()
    temp_dir = Path(tmpdir)
    Repo.clone_from("https://github.com/cruft/cookiecutter-test", temp_dir)

    # Create something deeper in the git tree
    project_dir = cruft.create(
        "https://github.com/cruft/cookiecutter-test",
        Path("tmpdir/foo/bar"),
        directory="dir",
        checkout="main",
    )
    # not added & committed
    assert not cruft.update(project_dir)
    # Add & commit the changes so that the repo is clean
    run(["git", "add", "."], cwd=temp_dir)
    run(
        [
            "git",
            "-c",
            "user.name='test'",
            "-c",
            "user.email='user@test.com'",
            "commit",
            "-am",
            "test",
        ],
        cwd=temp_dir,
    )

    assert cruft.update(project_dir, checkout="updated")



import json
from functools import partial
from pathlib import Path
from subprocess import run  # nosec
from textwrap import dedent

import pytest
from typer.testing import CliRunner

import cruft
from cruft._cli import app
from cruft._commands import utils


@pytest.fixture
def cruft_runner():
    runner = CliRunner()
    yield partial(runner.invoke, app)


@pytest.fixture
def cookiecutter_dir(tmpdir):
    yield Path(
        cruft.create("https://github.com/cruft/cookiecutter-test", Path(tmpdir), directory="dir")
    )


@pytest.fixture
def cookiecutter_dir_updated(tmpdir):
    yield Path(
        cruft.create(
            "https://github.com/cruft/cookiecutter-test",
            Path(tmpdir),
            directory="dir",
            checkout="updated",
        )
    )


@pytest.fixture
def cookiecutter_dir_hooked_git(tmpdir):
    yield Path(
        cruft.create(
            # See pull request!
            "https://github.com/cruft/cookiecutter-test",
            Path(tmpdir),
            directory="dir",
            checkout="with-git-from-hook",
        )
    )


@pytest.fixture
def cookiecutter_dir_input(tmpdir):
    yield Path(
        cruft.create(
            "https://github.com/cruft/cookiecutter-test",
            Path(tmpdir),
            directory="dir",
            checkout="input",
        )
    )


@pytest.fixture
def cookiecutter_dir_submodule(tmpdir):
    yield Path(
        cruft.create(
            "https://github.com/cruft/cookiecutter-test",
            Path(tmpdir),
            checkout="submodules",
        )
    )


@pytest.fixture
def cookiecutter_dir_extensions(tmpdir):
    yield Path(
        cruft.create(
            "https://github.com/cruft/cookiecutter-test",
            Path(tmpdir),
            checkout="extensions",
            directory="dir",
        )
    )


@pytest.fixture
def cookiecutter_no_dir_extensions(tmpdir):
    yield Path(
        cruft.create(
            "https://github.com/cruft/cookiecutter-test",
            Path(tmpdir),
            checkout="no-dir-extensions",
        )
    )


def test_create(cruft_runner, tmpdir):
    result = cruft_runner(
        [
            "create",
            "--output-dir",
            str(tmpdir),
            "https://github.com/cruft/cookiecutter-test",
            "--directory",
            "dir",
            "-y",
        ]
    )
    assert result.exit_code == 0
    assert result.stdout == ""


def test_create_interactive(cruft_runner, tmpdir):
    result = cruft_runner(
        [
            "create",
            "--output-dir",
            str(tmpdir),
            "https://github.com/cruft/cookiecutter-test",
            "--directory",
            "dir",
        ],
        input="RANDOM_NAME\n",
    )
    assert result.exit_code == 0
    assert (Path(tmpdir) / "RANDOM_NAME").exists()


def test_create_extra_context_file(cruft_runner, tmpdir):
    extra_context_file = Path(__file__).parent / "testdata" / "unicode-data" / "extra_context.json"
    result = cruft_runner(
        [
            "create",
            "--output-dir",
            str(tmpdir),
            "--extra-context-file",
            str(extra_context_file),
            "--directory",
            "dir",
            "--no-input",
            "https://github.com/cruft/cookiecutter-test",
        ]
    )
    assert result.exit_code == 0
    assert (Path(tmpdir) / "CRUFT-TEST-PROJECT").exists()


def test_check(cruft_runner, cookiecutter_dir):
    result = cruft_runner(["check", "--project-dir", cookiecutter_dir.as_posix()])
    assert result.exit_code == 0


def test_check_strict(cruft_runner, cookiecutter_dir_updated):
    result = cruft_runner(["check", "--project-dir", cookiecutter_dir_updated.as_posix()])
    assert result.exit_code == 1
    assert "failure" in result.stdout.lower()


def test_check_not_strict(cruft_runner, cookiecutter_dir_updated):
    result = cruft_runner(
        ["check", "--project-dir", cookiecutter_dir_updated.as_posix(), "--not-strict"]
    )
    assert result.exit_code == 0


def test_check_stale(cruft_runner, cookiecutter_dir):
    result = cruft_runner(
        ["check", "--project-dir", cookiecutter_dir.as_posix(), "--checkout", "updated"]
    )
    assert result.exit_code == 1
    assert "failure" in result.stdout.lower()


def test_link(cruft_runner, cookiecutter_dir):
    cruft_file = utils.cruft.get_cruft_file(cookiecutter_dir)
    cruft_config_from_create = json.loads(cruft_file.read_text())
    cruft_file.unlink()
    result = cruft_runner(
        [
            "link",
            "https://github.com/cruft/cookiecutter-test",
            "--project-dir",
            cookiecutter_dir.as_posix(),
            "-y",
            "--directory",
            "dir",
        ]
    )
    assert result.stdout == ""
    assert result.exit_code == 0

    # compare the 2 .cruft.json
    cruft_file = utils.cruft.get_cruft_file(cookiecutter_dir)
    cruft_config_from_link = json.loads(cruft_file.read_text())
    assert cruft_config_from_create == cruft_config_from_link


def test_link_interactive(cruft_runner, cookiecutter_dir):
    cruft_file = utils.cruft.get_cruft_file(cookiecutter_dir)
    cruft_config_from_create = json.loads(cruft_file.read_text())
    commit = cruft_config_from_create["commit"]
    assert commit == cruft_config_from_create["context"]["cookiecutter"]["_commit"]
    cruft_file.unlink()
    result = cruft_runner(
        [
            "link",
            "https://github.com/cruft/cookiecutter-test",
            "--project-dir",
            cookiecutter_dir.as_posix(),
            "--directory",
            "dir",
        ],
        input=f"{commit}\n",
    )
    assert "Link to template at commit" in result.stdout
    assert result.exit_code == 0

    # compare the 2 .cruft.json (except for the "project" key)
    cruft_file = utils.cruft.get_cruft_file(cookiecutter_dir)
    cruft_config_from_link = json.loads(cruft_file.read_text())
    cruft_config_from_create["context"]["cookiecutter"].pop("project")
    cruft_config_from_link["context"]["cookiecutter"].pop("project")
    assert cruft_config_from_create == cruft_config_from_link


def test_link_checkout(cruft_runner, cookiecutter_dir_updated):
    cruft_file = utils.cruft.get_cruft_file(cookiecutter_dir_updated)
    cruft_config_from_create = json.loads(cruft_file.read_text())
    cruft_file.unlink()
    result = cruft_runner(
        [
            "link",
            "https://github.com/cruft/cookiecutter-test",
            "--project-dir",
            cookiecutter_dir_updated.as_posix(),
            "-y",
            "--directory",
            "dir",
            "--checkout",
            "updated",
        ]
    )
    assert result.exit_code == 0

    # compare the 2 .cruft.json
    cruft_file = utils.cruft.get_cruft_file(cookiecutter_dir_updated)
    cruft_config_from_link = json.loads(cruft_file.read_text())
    assert cruft_config_from_create == cruft_config_from_link


def test_update_noop(cruft_runner, cookiecutter_dir):
    result = cruft_runner(["update", "--project-dir", cookiecutter_dir.as_posix(), "-y"])
    assert "Nothing to do" in result.stdout
    assert result.exit_code == 0


def test_update_unclean(cruft_runner, cookiecutter_dir):
    run(["git", "init"], cwd=cookiecutter_dir)
    result = cruft_runner(["update", "--project-dir", cookiecutter_dir.as_posix(), "-y"])
    assert "Cruft cannot apply updates on an unclean git project." in result.stdout
    assert result.exit_code == 1


def test_update_allow_untracked_files(cruft_runner, cookiecutter_dir):
    run(["git", "init"], cwd=cookiecutter_dir)
    run(["git", "add", "-A"], cwd=cookiecutter_dir)
    run(
        [
            "git",
            "-c",
            "user.name='test'",
            "-c",
            "user.email='user@test.com'",
            "commit",
            "-am",
            "test",
        ],
        cwd=cookiecutter_dir,
    )
    (cookiecutter_dir / "new_file.txt").touch()
    result = cruft_runner(["update", "--project-dir", str(cookiecutter_dir), "-y"])
    assert "Cruft cannot apply updates on an unclean git project." in result.stdout
    assert result.exit_code == 1
    result = cruft_runner(
        [
            "update",
            "--project-dir",
            str(cookiecutter_dir),
            "-y",
            "--allow-untracked-files",
            "-c",
            "updated",
        ]
    )
    assert "cruft has been updated" in result.stdout
    assert result.exit_code == 0


def test_update(cruft_runner, cookiecutter_dir):
    result = cruft_runner(
        ["update", "--project-dir", cookiecutter_dir.as_posix(), "-y", "-c", "updated"]
    )
    assert "cruft has been updated" in result.stdout
    assert result.exit_code == 0


def test_update_with_conflicts(cruft_runner, cookiecutter_dir):
    with (cookiecutter_dir / "README.md").open("w") as f:
        f.write("conflicts")
    result = cruft_runner(
        ["update", "--project-dir", cookiecutter_dir.as_posix(), "-y", "-c", "updated"]
    )
    assert "cruft has been updated" in result.stdout
    assert "Project directory may have *.rej files" in result.stdout
    assert result.exit_code == 0
    assert set(cookiecutter_dir.glob("**/*.rej"))


def test_update_with_conflicts_with_git(cruft_runner, cookiecutter_dir):
    with (cookiecutter_dir / "README.md").open("w") as f:
        f.write("conflicts")
    # Commit the changes so that the repo is clean
    run(["git", "init"], cwd=cookiecutter_dir)
    run(["git", "add", "-A"], cwd=cookiecutter_dir)
    run(
        [
            "git",
            "-c",
            "user.name='test'",
            "-c",
            "user.email='user@test.com'",
            "commit",
            "-am",
            "test",
        ],
        cwd=cookiecutter_dir,
    )
    result = cruft_runner(
        ["update", "--project-dir", cookiecutter_dir.as_posix(), "-y", "-c", "updated"]
    )
    assert "cruft has been updated" in result.stdout
    assert result.exit_code == 0
    assert set(cookiecutter_dir.glob("**/*.rej"))
    assert "Project directory may have *.rej files" in result.stdout
    assert "Retrying again with a different update strategy." in result.stdout


def test_update_interactive_cancelled(cruft_runner, cookiecutter_dir):
    result = cruft_runner(
        ["update", "--project-dir", cookiecutter_dir.as_posix(), "-c", "updated"], input="n\n"
    )
    assert result.exit_code == 0
    assert "User cancelled Cookiecutter template update" in result.stdout


def test_update_interactive_skip(cruft_runner, cookiecutter_dir):
    result = cruft_runner(
        ["update", "--project-dir", cookiecutter_dir.as_posix(), "-c", "updated"], input="s\n"
    )
    assert result.exit_code == 0
    assert "cruft has been updated" in result.stdout


def test_update_interactive_view(cruft_runner, cookiecutter_dir):
    result = cruft_runner(
        ["update", "--project-dir", cookiecutter_dir.as_posix(), "-c", "updated"], input="v\ny\n"
    )
    assert result.exit_code == 0
    assert "cruft has been updated" in result.stdout


def test_update_not_strict(cruft_runner, cookiecutter_dir_updated):
    result = cruft_runner(
        ["update", "--project-dir", cookiecutter_dir_updated.as_posix(), "--not-strict"]
    )
    assert result.exit_code == 0
    assert "Nothing to do, project's cruft is already up to date" in result.stdout


def test_update_strict(cruft_runner, cookiecutter_dir_updated):
    result = cruft_runner(["update", "--project-dir", cookiecutter_dir_updated.as_posix(), "-y"])
    assert result.exit_code == 0
    assert "cruft has been updated" in result.stdout


def test_update_when_new_file(cruft_runner, cookiecutter_dir):
    result = cruft_runner(
        ["update", "--project-dir", cookiecutter_dir.as_posix(), "-c", "new-file"], input="y\n"
    )
    assert result.exit_code == 0
    assert (cookiecutter_dir / "new-file").exists()
    assert "cruft has been updated" in result.stdout


def test_update_when_file_moved(cruft_runner, cookiecutter_dir):
    result = cruft_runner(
        ["update", "--project-dir", cookiecutter_dir.as_posix(), "-c", "file-moved"], input="y\n"
    )
    assert result.exit_code == 0
    assert (cookiecutter_dir / "NEW-README.md").exists()
    assert not (cookiecutter_dir / "README.md").exists()
    assert "cruft has been updated" in result.stdout


def test_update_interactive_view_no_changes(cruft_runner, cookiecutter_dir):
    result = cruft_runner(
        ["update", "--project-dir", cookiecutter_dir.as_posix(), "-c", "no-changes"], input="v\ny\n"
    )
    assert result.exit_code == 0
    assert "There are no changes" in result.stdout
    assert "cruft has been updated" in result.stdout


def test_update_interactive_view_no_changes_when_deleted(cruft_runner, cookiecutter_dir):
    # Remove the file that changed.
    (cookiecutter_dir / "README.md").unlink()
    result = cruft_runner(
        ["update", "--project-dir", cookiecutter_dir.as_posix(), "-c", "updated"], input="v\ny\n"
    )
    assert result.exit_code == 0
    assert "There are no changes" in result.stdout
    assert "cruft has been updated" in result.stdout


def test_update_same_commit_but_ask_for_input(cruft_runner, cookiecutter_dir_input):
    result = cruft_runner(
        ["update", "--project-dir", cookiecutter_dir_input.as_posix(), "-c", "input", "-y", "-i"],
        input="\n\n",  # no input changes
    )
    assert "cruft has been updated" in result.stdout
    assert result.exit_code == 0


def test_update_with_input_changes(cruft_runner, cookiecutter_dir_input, capfd):
    result = cruft_runner(
        ["update", "--project-dir", cookiecutter_dir_input.as_posix(), "-c", "input", "-i"],
        input="test\nnew-input\nv\ny\n",
    )
    git_diff_captured = capfd.readouterr()
    assert "-Input from cookiecutter: some-input" in git_diff_captured.out
    assert "+Input from cookiecutter: new-input" in git_diff_captured.out
    assert "cruft has been updated" in result.stdout
    assert result.exit_code == 0


def test_update_new_inputs_added_to_template(cruft_runner, cookiecutter_dir_input, capfd):
    result = cruft_runner(
        ["update", "--project-dir", cookiecutter_dir_input.as_posix(), "-c", "input-updated", "-i"],
        input="test\nsome-input\nnew-input-from-template\nv\ny\n",
    )
    git_diff_captured = capfd.readouterr()
    assert "-Initial" in git_diff_captured.out
    assert "+Updated" in git_diff_captured.out
    assert "+New input added from template: new-input-from-template" in git_diff_captured.out
    assert "cruft has been updated" in result.stdout
    assert result.exit_code == 0


def test_update_refresh_private_variables_from_template(
    cruft_runner, cookiecutter_dir_input, capfd
):
    result = cruft_runner(
        ["update", "--project-dir", cookiecutter_dir_input.as_posix(), "-c", "input-updated", "-r"],
        input="v\ny\n",
    )
    git_diff_captured = capfd.readouterr()
    assert "-Private variable: 1.0" in git_diff_captured.out
    assert "+Private variable: 2.0" in git_diff_captured.out
    assert "cruft has been updated" in result.stdout
    assert result.exit_code == 0


def make_cruft_file_content(context):
    return {
        "template": "https://github.com/gmsantos/cookiecutter-test",
        "commit": "ab2e492522eddb65c44c2848d6de074ba977b1ac",
        "checkout": "input",
        "context": {
            "cookiecutter": context,
        },
        "directory": "dir",
    }


@pytest.mark.parametrize(
    "template_version",
    [
        pytest.param("input", id="version:initial"),
        pytest.param("input-updated", id="version:updated"),
    ],
)
@pytest.mark.parametrize(
    "variables_to_update_cli",
    [
        pytest.param((None, None, False), id="cli:none"),
        pytest.param(("{}", None, False), id="cli:empty"),
        pytest.param(('{"input":"new-cli-value"}', "new-cli-value", True), id="cli:single1"),
        pytest.param(('{"input":"other-cli-value"}', "other-cli-value", True), id="cli:single2"),
        pytest.param(('{"not-an-input":"new-cli-value"}', None, True), id="cli:unsused-var"),
        pytest.param(
            ('{"not-an-input":"not-used","input":"new-cli-value"}', "new-cli-value", True),
            id="cli:multiple",
        ),
    ],
)
@pytest.mark.parametrize(
    "variables_to_update_file",
    [
        pytest.param((None, None, False), id="file:none"),
        pytest.param(({}, None, False), id="file:empty-level0"),
        pytest.param(
            (
                {
                    "template": "https://github.com/gmsantos/cookiecutter-test",
                    "commit": "ab2e492522eddb65c44c2848d6de074ba977b1ac",
                    "checkout": "input",
                    "directory": "dir",
                },
                None,
                False,
            ),
            id="file:empty-level1",
        ),
        pytest.param((make_cruft_file_content({}), None, False), id="file:empty-level2"),
        pytest.param(
            (
                make_cruft_file_content({"input": "new-file-value"}),
                "new-file-value",
                True,
            ),
            id="file:single1",
        ),
        pytest.param(
            (
                make_cruft_file_content({"input": "other-file-value"}),
                "other-file-value",
                True,
            ),
            id="file:single2",
        ),
        pytest.param(
            (
                make_cruft_file_content({"not-used": "new-file-value"}),
                None,
                True,
            ),
            id="file:unused-var",
        ),
        pytest.param(
            (
                make_cruft_file_content(
                    {
                        "input": "other-file-value",
                        "ignored": "not-used",
                    }
                ),
                "other-file-value",
                True,
            ),
            id="file:single2",
        ),
    ],
)
def test_update_changed_variables(
    template_version,
    variables_to_update_cli,
    variables_to_update_file,
    cruft_runner,
    cookiecutter_dir_input,
    capfd,
    tmp_path,
):
    vtu_cli_input, vtu_cli_expected, vtu_cli_updates = variables_to_update_cli
    vtu_file_input, vtu_file_expected, vtu_file_updates = variables_to_update_file

    expecting_updates = vtu_cli_updates or vtu_file_updates

    cmd_args = [
        "update",
        "--project-dir",
        cookiecutter_dir_input.as_posix(),
        "-c",
        template_version,
    ]

    if vtu_cli_input:
        cmd_args.append(f"--variables-to-update={vtu_cli_input}")

    if vtu_file_input:
        variables_to_update_file = tmp_path / "new-cruft-file.json"
        with open(variables_to_update_file, "w", encoding="utf-8") as fd:
            json.dump(vtu_file_input, fd)
        cmd_args.append(f"--variables-to-update-file={variables_to_update_file}")

    result = cruft_runner(cmd_args, input="v\ny\n")

    git_diff_captured = capfd.readouterr()

    expected_input_value = vtu_cli_expected or vtu_file_expected  # CLI takes precedence

    # validate input value used in project
    if expected_input_value:
        assert f"+Input from cookiecutter: {expected_input_value}" in git_diff_captured.out
        assert "-Input from cookiecutter: some-input" in git_diff_captured.out
    else:
        assert "-Input from cookiecutter:" not in git_diff_captured.out
        assert "+Input from cookiecutter:" not in git_diff_captured.out

    # validate template versions
    if template_version == "input-updated":
        expecting_updates = True
        assert "-Initial" in git_diff_captured.out
        assert "+Updated" in git_diff_captured.out

    # validate overall result
    if expecting_updates:
        assert "cruft has been updated" in result.stdout
    else:
        assert "cruft has been updated" not in result.stdout
        assert "already up to date" in result.stdout
    assert result.exit_code == 0


def test_update_changed_variables_wrong_file(
    cruft_runner,
    cookiecutter_dir_input,
    capfd,
):
    result = cruft_runner(
        [
            "update",
            "--project-dir",
            cookiecutter_dir_input.as_posix(),
            "-c",
            "input",
            f"--variables-to-update-file={cookiecutter_dir_input.as_posix()}/.cruft.json",
        ],
        input="v\ny\n",
    )

    capfd.readouterr()

    assert "cannot be the same as the project's cruft file" in result.stdout
    assert result.exit_code != 0


@pytest.mark.parametrize("args,expected_exit_code", [([], 0), (["--exit-code"], 1), (["-e"], 1)])
def test_diff_has_diff(args, expected_exit_code, cruft_runner, cookiecutter_dir):
    (cookiecutter_dir / "README.md").write_text("changed content\n")
    result = cruft_runner(["diff", "--project-dir", cookiecutter_dir.as_posix()] + args)
    assert result.exit_code == expected_exit_code
    assert result.stdout != ""


def test_diff_checkout(cruft_runner, cookiecutter_dir):
    result = cruft_runner(
        [
            "diff",
            "--project-dir",
            cookiecutter_dir.as_posix(),
            "--checkout",
            "updated",
            "--exit-code",
        ]
    )
    assert result.exit_code == 1
    assert result.stdout != ""


@pytest.mark.parametrize("args,expected_exit_code", [([], 0), (["--exit-code"], 0), (["-e"], 0)])
def test_diff_no_diff(args, expected_exit_code, cruft_runner, cookiecutter_dir):
    result = cruft_runner(["diff", "--project-dir", cookiecutter_dir.as_posix()] + args)
    assert result.exit_code == expected_exit_code
    assert result.stdout == ""


@pytest.mark.parametrize("args, expected_exit_code", [([], 0)])
def test_diff_skip_git_dir(args, expected_exit_code, cruft_runner, cookiecutter_dir_hooked_git):
    cookiecutter_dir = cookiecutter_dir_hooked_git
    # The two points below could as well be nicely stored within
    # a cookiecutter-test branch.
    # Write a skip section into pyproject.toml
    # This file is not (yet) in the test branch and is thus not diffed to the template.
    skip_section = """
        [tool.cruft]
        skip = [".git"]
        """
    with (cookiecutter_dir / "pyproject.toml").open("w") as f:
        f.write(dedent(skip_section))
    # Alter the git repo.
    # run(["git", "config", "--global", "user.email", "user@test.com"], cwd=cookiecutter_dir)
    # run(["git", "config", "--global", "user.name", "testm"], cwd=cookiecutter_dir)
    run(["git", "add", "--all"], cwd=cookiecutter_dir)
    run(["git", "commit", "-m", "2nd commit"], cwd=cookiecutter_dir)
    result = cruft_runner(["diff", "--project-dir", cookiecutter_dir.as_posix(), "--exit-code"])
    assert result.exit_code == expected_exit_code
    assert ".git" not in result.stdout


def test_local_extension_check(cruft_runner, cookiecutter_dir_extensions):
    result = cruft_runner(
        [
            "check",
            "--project-dir",
            str(cookiecutter_dir_extensions),
            "--checkout",
            "extensions-update",
        ]
    )
    assert result.exit_code == 1
    assert (
        "Project's cruft is out of date! Run `cruft update` to clean this mess up." in result.stdout
    )


def test_local_extension_diff(cruft_runner, cookiecutter_dir_extensions):
    result = cruft_runner(
        [
            "diff",
            "--project-dir",
            str(cookiecutter_dir_extensions),
            "--checkout",
            "extensions-update",
        ]
    )
    assert result.exit_code == 0
    assert (
        "diff --git upstream-template-old/README.md upstream-template-new/README.md"
        in result.stdout
    )


def test_local_extension_update(cruft_runner, cookiecutter_dir_extensions):
    result = cruft_runner(
        [
            "update",
            "--project-dir",
            str(cookiecutter_dir_extensions),
            "--checkout",
            "extensions-update",
            "--skip-apply-ask",
        ]
    )
    assert result.exit_code == 0
    with open(cookiecutter_dir_extensions / "README.md") as f:
        assert "Updated11" in f.read()


def test_local_extension_without_dir(cruft_runner, cookiecutter_no_dir_extensions):
    result = cruft_runner(
        [
            "create",
            "--output-dir",
            str(cookiecutter_no_dir_extensions),
            "https://github.com/cruft/cookiecutter-test",
            "--checkout",
            "no-dir-extensions",
            "-y",
        ]
    )
    assert result.exit_code == 0
    assert result.stdout == ""


def test_local_extension_without_dir_check(cruft_runner, cookiecutter_no_dir_extensions):
    result = cruft_runner(
        [
            "check",
            "--project-dir",
            str(cookiecutter_no_dir_extensions),
            "--checkout",
            "no-dir-extensions-update",
        ]
    )
    assert result.exit_code == 1
    assert (
        "Project's cruft is out of date! Run `cruft update` to clean this mess up." in result.stdout
    )


def test_local_extension_without_dir_diff(cruft_runner, cookiecutter_no_dir_extensions):
    result = cruft_runner(
        [
            "diff",
            "--project-dir",
            str(cookiecutter_no_dir_extensions),
            "--checkout",
            "no-dir-extensions-update",
        ]
    )
    assert result.exit_code == 0
    assert (
        "diff --git upstream-template-old/README.md upstream-template-new/README.md"
        in result.stdout
    )


def test_local_extension_without_dir_update(cruft_runner, cookiecutter_no_dir_extensions):
    result = cruft_runner(
        [
            "update",
            "--project-dir",
            str(cookiecutter_no_dir_extensions),
            "--checkout",
            "no-dir-extensions-update",
            "--skip-apply-ask",
        ]
    )
    assert result.exit_code == 0
    with open(cookiecutter_no_dir_extensions / "README.md") as f:
        assert "Updated11" in f.read()


def test_submodule_create(cruft_runner, cookiecutter_dir_submodule):
    # the submodule was properly cloned if the file of the submodule exists
    assert (cookiecutter_dir_submodule / "submodule" / "test-file").exists()


def test_submodule_update_has_submodule_diff(cruft_runner, cookiecutter_dir_submodule, capfd):
    # the diff during an update should include the submodule changes
    result = cruft_runner(
        [
            "update",
            "--project-dir",
            cookiecutter_dir_submodule.as_posix(),
            "-c",
            "submodules-updated",
        ],
        input="v\ny\n",
    )
    assert result.exit_code == 0

    git_diff_captured = capfd.readouterr()

    assert "current_template/submodule/test-file" in git_diff_captured.out
    assert "new_template/submodule/test-file" in git_diff_captured.out
    assert "@@ -1 +1 @@" in git_diff_captured.out
    assert "-revision 1" in git_diff_captured.out
    assert "+revision 2" in git_diff_captured.out

    assert "cruft has been updated" in result.stdout


def test_submodule_diff_includes_submodule(cruft_runner, cookiecutter_dir_submodule):
    with open(cookiecutter_dir_submodule / "submodule" / "test-file", "w") as f:
        f.write("revision 3")

    result = cruft_runner(["diff", "--project-dir", cookiecutter_dir_submodule.as_posix()])
    assert result.exit_code == 0

    assert "@@ -1 +1 @@" in result.stdout
    assert "-revision 3" in result.stdout
    assert "+revision 1" in result.stdout
